@@ -12,6 +12,8 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+#![allow(clippy::too_many_arguments)]
+
 pub use cl3::command_queue::*;
 
 use super::context::Context;
@@ -34,6 +36,7 @@ use cl3::ext;
 #[allow(unused_imports)]
 use cl3::ffi::cl_ext::cl_queue_properties_khr;
 use cl3::gl;
+#[allow(unused_imports)]
 use cl3::types::{
     cl_bool, cl_command_queue, cl_command_queue_properties, cl_context, cl_device_id, cl_event,
     cl_kernel, cl_map_flags, cl_mem, cl_mem_migration_flags, cl_queue_properties, cl_uint,
@@ -112,6 +115,7 @@ impl CommandQueue {
     ///
     /// returns a Result containing the new CommandQueue
     /// or the error code from the OpenCL C API function.
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn create_with_properties(
         context: &Context,
         device_id: cl_device_id,
@@ -187,7 +191,7 @@ impl CommandQueue {
             (data.len() * mem::size_of::<T>()) as size_t,
             data.as_mut_ptr() as cl_mem,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -223,7 +227,7 @@ impl CommandQueue {
             host_slice_pitch,
             ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -234,7 +238,7 @@ impl CommandQueue {
 
     pub fn enqueue_write_buffer<T>(
         &self,
-        buffer: &Buffer<T>,
+        buffer: &mut Buffer<T>,
         blocking_write: cl_bool,
         offset: size_t,
         data: &[T],
@@ -242,13 +246,13 @@ impl CommandQueue {
     ) -> Result<Event> {
         let event = enqueue_write_buffer(
             self.queue,
-            buffer.get(),
+            buffer.get_mut(),
             blocking_write,
             offset,
             (data.len() * mem::size_of::<T>()) as size_t,
             data.as_ptr() as cl_mem,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -259,7 +263,7 @@ impl CommandQueue {
 
     pub fn enqueue_write_buffer_rect<T>(
         &self,
-        buffer: &Buffer<T>,
+        buffer: &mut Buffer<T>,
         blocking_write: cl_bool,
         buffer_origin: *const size_t,
         host_origin: *const size_t,
@@ -273,7 +277,7 @@ impl CommandQueue {
     ) -> Result<Event> {
         let event = enqueue_write_buffer_rect(
             self.queue,
-            buffer.get(),
+            buffer.get_mut(),
             blocking_write,
             buffer_origin,
             host_origin,
@@ -284,7 +288,7 @@ impl CommandQueue {
             host_slice_pitch,
             ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -293,9 +297,10 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_fill_buffer<T>(
         &self,
-        buffer: &Buffer<T>,
+        buffer: &mut Buffer<T>,
         pattern: &[T],
         offset: size_t,
         size: size_t,
@@ -303,13 +308,13 @@ impl CommandQueue {
     ) -> Result<Event> {
         let event = enqueue_fill_buffer(
             self.queue,
-            buffer.get(),
+            buffer.get_mut(),
             pattern.as_ptr() as cl_mem,
             pattern.len() * mem::size_of::<T>(),
             offset,
             size,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -321,7 +326,7 @@ impl CommandQueue {
     pub fn enqueue_copy_buffer<T>(
         &self,
         src_buffer: &Buffer<T>,
-        dst_buffer: &Buffer<T>,
+        dst_buffer: &mut Buffer<T>,
         src_offset: size_t,
         dst_offset: size_t,
         size: size_t,
@@ -330,12 +335,12 @@ impl CommandQueue {
         let event = enqueue_copy_buffer(
             self.queue,
             src_buffer.get(),
-            dst_buffer.get(),
+            dst_buffer.get_mut(),
             src_offset,
             dst_offset,
             size,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -346,7 +351,7 @@ impl CommandQueue {
     pub fn enqueue_copy_buffer_rect<T>(
         &self,
         src_buffer: &Buffer<T>,
-        dst_buffer: &Buffer<T>,
+        dst_buffer: &mut Buffer<T>,
         src_origin: *const size_t,
         dst_origin: *const size_t,
         region: *const size_t,
@@ -359,7 +364,7 @@ impl CommandQueue {
         let event = enqueue_copy_buffer_rect(
             self.queue,
             src_buffer.get(),
-            dst_buffer.get(),
+            dst_buffer.get_mut(),
             src_origin,
             dst_origin,
             region,
@@ -368,7 +373,7 @@ impl CommandQueue {
             dst_row_pitch,
             dst_slice_pitch,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -398,7 +403,7 @@ impl CommandQueue {
             slice_pitch,
             ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -409,7 +414,7 @@ impl CommandQueue {
 
     pub fn enqueue_write_image(
         &self,
-        image: &Image,
+        image: &mut Image,
         blocking_write: cl_bool,
         origin: *const size_t,
         region: *const size_t,
@@ -420,7 +425,7 @@ impl CommandQueue {
     ) -> Result<Event> {
         let event = enqueue_write_image(
             self.queue,
-            image.get(),
+            image.get_mut(),
             blocking_write,
             origin,
             region,
@@ -428,7 +433,7 @@ impl CommandQueue {
             slice_pitch,
             ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -437,9 +442,10 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_fill_image(
         &self,
-        image: &Image,
+        image: &mut Image,
         fill_color: *const c_void,
         origin: *const size_t,
         region: *const size_t,
@@ -447,12 +453,12 @@ impl CommandQueue {
     ) -> Result<Event> {
         let event = enqueue_fill_image(
             self.queue,
-            image.get(),
+            image.get_mut(),
             fill_color,
             origin,
             region,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -464,7 +470,7 @@ impl CommandQueue {
     pub fn enqueue_copy_image(
         &self,
         src_image: &Image,
-        dst_image: &Image,
+        dst_image: &mut Image,
         src_origin: *const size_t,
         dst_origin: *const size_t,
         region: *const size_t,
@@ -473,12 +479,12 @@ impl CommandQueue {
         let event = enqueue_copy_image(
             self.queue,
             src_image.get(),
-            dst_image.get(),
+            dst_image.get_mut(),
             src_origin,
             dst_origin,
             region,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -490,7 +496,7 @@ impl CommandQueue {
     pub fn enqueue_copy_image_to_buffer<T>(
         &self,
         src_image: &Image,
-        dst_buffer: &Buffer<T>,
+        dst_buffer: &mut Buffer<T>,
         src_origin: *const size_t,
         region: *const size_t,
         dst_offset: size_t,
@@ -499,12 +505,12 @@ impl CommandQueue {
         let event = enqueue_copy_image_to_buffer(
             self.queue,
             src_image.get(),
-            dst_buffer.get(),
+            dst_buffer.get_mut(),
             src_origin,
             region,
             dst_offset,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -516,7 +522,7 @@ impl CommandQueue {
     pub fn enqueue_copy_buffer_to_image<T>(
         &self,
         src_buffer: &Buffer<T>,
-        dst_image: &Image,
+        dst_image: &mut Image,
         src_offset: size_t,
         dst_origin: *const size_t,
         region: *const size_t,
@@ -525,12 +531,12 @@ impl CommandQueue {
         let event = enqueue_copy_buffer_to_image(
             self.queue,
             src_buffer.get(),
-            dst_image.get(),
+            dst_image.get_mut(),
             src_offset,
             dst_origin,
             region,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -558,7 +564,7 @@ impl CommandQueue {
             size,
             buffer_ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -590,7 +596,7 @@ impl CommandQueue {
             image_slice_pitch,
             image_ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -610,7 +616,7 @@ impl CommandQueue {
             memobj,
             mapped_ptr,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -619,6 +625,7 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_migrate_mem_object(
         &self,
         num_mem_objects: cl_uint,
@@ -632,7 +639,7 @@ impl CommandQueue {
             mem_objects,
             flags,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -655,7 +662,7 @@ impl CommandQueue {
             mem_objects,
             flags,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -681,7 +688,7 @@ impl CommandQueue {
             global_work_sizes,
             local_work_sizes,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -690,12 +697,13 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_task(&self, kernel: cl_kernel, event_wait_list: &[cl_event]) -> Result<Event> {
         let event = enqueue_task(
             self.queue,
             kernel,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -718,14 +726,14 @@ impl CommandQueue {
             args.as_ptr() as *mut c_void,
             args.len() as size_t,
             mem_list.len() as cl_uint,
-            if 0 < mem_list.len() {
+            if !mem_list.is_empty() {
                 mem_list.as_ptr()
             } else {
                 ptr::null()
             },
             args_mem_loc.as_ptr(),
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -734,11 +742,12 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_marker_with_wait_list(&self, event_wait_list: &[cl_event]) -> Result<Event> {
         let event = enqueue_marker_with_wait_list(
             self.queue,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -747,11 +756,12 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_1_2")]
     pub fn enqueue_barrier_with_wait_list(&self, event_wait_list: &[cl_event]) -> Result<Event> {
         let event = enqueue_barrier_with_wait_list(
             self.queue,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -760,6 +770,7 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn enqueue_svm_free(
         &self,
         svm_pointers: &[*const c_void],
@@ -781,7 +792,7 @@ impl CommandQueue {
             pfn_free_func,
             user_data,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -790,6 +801,7 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn enqueue_svm_mem_cpy(
         &self,
         blocking_copy: cl_bool,
@@ -805,7 +817,7 @@ impl CommandQueue {
             src_ptr,
             size,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -814,6 +826,7 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn enqueue_svm_mem_fill<T>(
         &self,
         svm_ptr: *mut c_void,
@@ -828,7 +841,7 @@ impl CommandQueue {
             pattern.len() * mem::size_of::<T>(),
             size,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -837,6 +850,7 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn enqueue_svm_map<T>(
         &self,
         blocking_map: cl_bool,
@@ -851,7 +865,7 @@ impl CommandQueue {
             svm.as_mut_ptr() as *mut c_void,
             svm.len() * mem::size_of::<T>(),
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -860,12 +874,13 @@ impl CommandQueue {
         Ok(Event::new(event))
     }
 
+    #[cfg(feature = "CL_VERSION_2_0")]
     pub fn enqueue_svm_unmap<T>(&self, svm: &[T], event_wait_list: &[cl_event]) -> Result<Event> {
         let event = enqueue_svm_unmap(
             self.queue,
             svm.as_ptr() as *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -889,7 +904,7 @@ impl CommandQueue {
             sizes,
             flags,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -908,7 +923,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -927,7 +942,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -948,7 +963,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -969,7 +984,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -990,7 +1005,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1011,7 +1026,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1032,7 +1047,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1053,7 +1068,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1074,7 +1089,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1095,7 +1110,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1116,7 +1131,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1137,7 +1152,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1158,7 +1173,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1179,7 +1194,7 @@ impl CommandQueue {
             mem_objects.len() as cl_uint,
             mem_objects.as_ptr() as *const *mut c_void,
             event_wait_list.len() as cl_uint,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1206,7 +1221,7 @@ impl CommandQueue {
             mipmap_filter_mode,
             array_region,
             mip_region,
-            if 0 < event_wait_list.len() {
+            if !event_wait_list.is_empty() {
                 event_wait_list.as_ptr()
             } else {
                 ptr::null()
@@ -1240,11 +1255,12 @@ impl CommandQueue {
         Ok(get_command_queue_info(self.queue, CommandQueueInfo::CL_QUEUE_PROPERTIES)?.to_ulong())
     }
 
+    /// CL_VERSION_2_0
     pub fn size(&self) -> Result<cl_uint> {
         Ok(get_command_queue_info(self.queue, CommandQueueInfo::CL_QUEUE_SIZE)?.to_uint())
     }
 
-    // CL_VERSION_2_1
+    /// CL_VERSION_2_1
     pub fn device_default(&self) -> Result<cl_device_id> {
         Ok(
             get_command_queue_info(self.queue, CommandQueueInfo::CL_QUEUE_DEVICE_DEFAULT)?.to_ptr()
@@ -1252,7 +1268,7 @@ impl CommandQueue {
         )
     }
 
-    // CL_VERSION_3_0
+    /// CL_VERSION_3_0
     pub fn properties_array(&self) -> Result<Vec<cl_ulong>> {
         Ok(
             get_command_queue_info(self.queue, CommandQueueInfo::CL_QUEUE_PROPERTIES_ARRAY)?
